# This (sub-)makefile is part of the Classic McEliece hardware source code.
# 
# Copyright (C) 2022
#
# Authors: Norman Lahr <norman@lahr.email>
# 
# This source code is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# This source code is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this source code. If not, see <https://www.gnu.org/licenses/>. 

# Define required tools
MAKE ?= make
MAKE_VERSION_MIN := 4.3
MAKE_VERSION := $(shell $(MAKE) --version | grep Make | cut -d" " -f3)
ifndef NO_CHECK
ifeq ($(shell printf '%s\n%s\n' $(MAKE_VERSION_MIN) $(MAKE_VERSION) | sort --check=quiet --version-sort; echo $$?), 1)
    $(error Make version >= $(MAKE_VERSION_MIN) required, $(MAKE_VERSION) installed.)
endif
endif

PYTHON ?= python3
PYTHON_VERSION_MIN := 3.8.0
PYTHON_VERSION := $(shell $(PYTHON) --version | cut -d" " -f2)
ifndef NO_CHECK
ifeq ($(shell printf '%s\n%s\n' $(PYTHON_VERSION_MIN) $(PYTHON_VERSION) | sort --check=quiet --version-sort; echo $$?), 1)
    $(error Python version >= $(PYTHON_VERSION_MIN) required, $(PYTHON_VERSION) installed.)
endif
endif

SAGE ?= sage
SAGE_VERSION_MIN := 9.3
SAGE_VERSION := $(shell $(SAGE) --version | grep SageMath | cut -d"," -f1 | cut -d" " -f3)
ifndef NO_CHECK
ifeq ($(shell printf '%s\n%s\n' $(SAGE_VERSION_MIN) $(SAGE_VERSION) | sort --check=quiet --version-sort; echo $$?), 1)
    $(error SageMath version >= $(SAGE_VERSION_MIN) required, $(SAGE_VERSION) installed.)
endif
endif

VERILATOR ?= verilator
VERILATOR_VERSION_MIN := 4.216
VERILATOR_VERSION := $(shell $(VERILATOR) --version | cut -d" " -f2)
ifndef NO_CHECK
ifeq ($(shell printf '%s\n%s\n' $(VERILATOR_VERSION_MIN) $(VERILATOR_VERSION) | sort --check=quiet --version-sort; echo $$?), 1)
    $(error Verilator version >= $(VERILATOR_VERSION_MIN) required, $(VERILATOR_VERSION) installed.)
endif
endif

VIVADO ?= vivado
VIVADO_VERSION_MIN := 2019.2.1
VIVADO_VERSION := $(shell $(VIVADO) -version | grep Vivado | cut -d" " -f2 | cut -d"v" -f2)
ifndef NO_CHECK
ifeq ($(shell printf '%s\n%s\n' $(VIVADO_VERSION_MIN) $(VIVADO_VERSION) | sort --check=quiet --version-sort; echo $$?), 1)
    $(error Vivado version >= $(VIVADO_VERSION_MIN) required, $(VIVADO_VERSION) installed.)
endif
endif
