# This (sub-)makefile is part of the Classic McEliece hardware source code.
# 
# Copyright (C) 2022
#
# Authors: Norman Lahr <norman@lahr.email>
# 
# This source code is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# This source code is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this source code. If not, see <https://www.gnu.org/licenses/>. 

GEN_H_ENTRIES_SRC_PATH := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

GEN_H_ENTRIES_SUBMODULES := LOG2 MEMORY_MEM GF_MUL

include $(COMMON_SRC_PATH)/math/module.mk
include $(COMMON_SRC_PATH)/memory/module.mk

GEN_H_ENTRIES_SRC = gen_H_entries.v vec_mul_multi_H_gen.v vec_mul_H_gen.v

GEN_H_ENTRIES_SRC += $(foreach module,$(GEN_H_ENTRIES_SUBMODULES),$($(module)_SRC))

GEN_H_ENTRIES =

gen_H_entries: $(GEN_H_ENTRIES)

# Specify targets for each parameter set
define GEN_H_ENTRIES_SRC_TEMPLATE =

# Define all required targets to be included in the modules above
GEN_H_ENTRIES_$(1) = $(addprefix $$(BUILD_DIR_SRC_$(1))/,$$(GEN_H_ENTRIES_SRC))

# Collect all targets
GEN_H_ENTRIES += GEN_H_ENTRIES_$(1)

$$(BUILD_DIR_SRC_$(1))/gen_H_entries.v: $$(GEN_H_ENTRIES_SRC_PATH)/gen_H_entries.v
	cp $$< $$@

$$(BUILD_DIR_SRC_$(1))/vec_mul_multi_H_gen.v: $(MATH_SRC_PATH)/gen_vec_mul_multi.py
	$$(PYTHON) $$< -n $$(mul_tot_$(1)) --sec $$(mul_cyc_$(1)) --gf $$(m_$(1)) --name vec_mul_multi_H_gen --name_sub vec_mul_H_gen > $$@

$$(BUILD_DIR_SRC_$(1))/vec_mul_H_gen.v: $(MATH_SRC_PATH)/gen_vec_mul.py
	$$(PYTHON) $$< -n $$(mul_cyc_$(1)) --gf $$(m_$(1)) --name vec_mul_H_gen > $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call GEN_H_ENTRIES_SRC_TEMPLATE,$(par))))
