/*
 * This file is the testbench file for elim.v.
 *s
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/


`timescale 1ns/1ps

module gen_H_entries_tb;

  //inputs
  reg clk = 0;
  reg start = 0;
  reg [`CLOG2(`K/`N)-1:0] block = 0;
  reg [`N*`m-1:0] index = 0;
  reg [`N*`m-1:0] eva_inv = 0;

  //outputs
  wire wr_en;
  wire [`N-1:0] data_in;
  wire [`CLOG2(`L*`K/`N)-1:0] wr_addr;
  wire block_init_done;

  gen_H_entries #(.N(`N), .m(`m), .K(`K), .L(`L)) DUT (
    .clk(clk),
    .start(start),
    .block(block),
    .index(index),
    .eva_inv(eva_inv),
    .wr_en(wr_en),
    .data_in(data_in),
    .wr_addr(wr_addr),
    .block_init_done(block_init_done)
  );
 
  reg [`CLOG2(`L*`K/`N)-1:0] rd_addr = 0;
  reg rd_en = 0;
  wire [`N-1:0] data_out;

  // M_20k for storing dout
  mem mem_data (
    .clock(clk),
    .data(data_in),
	  .rdaddress(rd_addr),
	  .rden(rd_en),
    .wraddress(wr_addr),
    .wren(wr_en),
	  .q(data_out)
  );

  defparam mem_data.WIDTH = `N;
  defparam mem_data.DEPTH = `L*`K/`N;
  defparam mem_data.FILE = "";

   
  initial
    begin
       $dumpfile("gen_H_entries_tb.vcd");
       $dumpvars(0, gen_H_entries_tb);
    end
 
  integer i;
  integer scan_file;
	
  integer STDERR = 32'h8000_0002;
  integer STDIN  = 32'h8000_0000;

  initial
    begin
       scan_file = $fscanf(STDIN, "%b\n", block);
       scan_file = $fscanf(STDIN, "%b\n", index);
       scan_file = $fscanf(STDIN, "%b\n", eva_inv);
    end
	
  integer start_time;
  
  initial
    begin
	rd_en <= 1'b0;
	rd_addr <= 0;
	# 25;
	start_time = $time;
	start <= 1;
	# 10;
	start <= 0;
	@(posedge block_init_done);
	 $fdisplay(STDERR, "\nruntime for init: %0d cycles\n", ($time-start_time)/10);
	@(negedge clk);
	rd_en <= 1'b1;
	for (i = block*`L; i < ((block+1)*`L); i = i+1)
		begin
			rd_addr <= i;
			# 10;
			$write("%b\n", data_out);
		end
	 $fdisplay(STDERR, "\ntotal runtime: %0d cycles\n", ($time-start_time)/10);  
	# 50;
	$finish;
				
	 end

  always 
    #5 clk = ! clk; 
   
endmodule
