#!/usr/bin/sage

#
# This file is part of the testbench, which generates a random input polynomial.
# 
# Copyright (C) 2018
# Authors: Wen Wang <wen.wang.ww349@yale.edu>
#          Ruben Niederhagen <ruben@polycephaly.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys

N = int(sys.argv[1])
L = int(sys.argv[2])
K = int(sys.argv[3])
m = int(sys.argv[4])

t = int(L/m)

fmt = "{0:0" + str(m) + "b}"

F.<x> = GF(2)[]
KK.<a> = GF(2^m, modulus="first_lexicographic")

data_in = []

for line in sys.stdin:
  data_in.append(line)

block_in = data_in[0]
index_in = data_in[1]
eva_in = data_in[2]

block = int(block_in[:m], 2)

index = []
alpha = []

while index_in != "\n":
  a = index_in[:m]
  index.append(int(a, 2))
  alpha.append(KK.fetch_int(int(a[::-1], 2)))
  index_in = index_in[m:]

eva = []

while eva_in != "\n":
  eva.append(KK.fetch_int(int(eva_in[:m], 2)))
  eva_in = eva_in[m:]

#print "eva = \n", eva

dout = matrix(KK, t, N)

# sequence: natural matrix view
for i in range(t):
  for j in range(N):
    dout[i, j] = eva[j]*(alpha[j])^i

# data needs to be stored from LSB
for i in range(t):
  for j in reversed(range(m)):
    for k in range(N):
      A = fmt.format(dout[i, k].integer_representation())
      sys.stdout.write(A[j])
    sys.stdout.write("\n")

 
  
