/*
 * This file is the testbench file for inv_g_alpha.v.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/


`timescale 1ns/1ps

module inv_g_alpha_tb;
  //inputs
  reg clk = 0;
  reg start = 0;
  wire [`m-1:0] perm_dout;
  wire [eva_mem_WIDTH-1:0] eva_dout;
  reg [`CLOG2(K_elim/`N)-1:0] block = 0;
  
  //outputs
  wire perm_rd_en;
  wire [`m-1:0] perm_rd_addr;
  wire eva_rd_en;
  wire [`CLOG2(`t+1)-1:0] eva_rd_addr;
  wire [`N*`m-1:0] index_out;
  wire [`N*`m-1:0] inv_out;
  wire done;
  
  localparam K_elim = `N*((`K+`N-1)/`N);
  localparam eva_mem_WIDTH = `m*(1 << (`m-`CLOG2(`t+1)));//memory width in the add_FFT module
  parameter eva_mem_DEPTH = (1 << `CLOG2(`t+1)); //memory depth in the add_FFT module
  
  inv_g_alpha #(.N(`N), .m(`m), .t(`t), .K(`K)) DUT (
    .clk(clk),
    .start(start),
    .perm_dout(perm_dout),
	  .eva_dout(eva_dout),
    .perm_rd_en(perm_rd_en),
    .perm_rd_addr(perm_rd_addr),
    .eva_rd_en(eva_rd_en),
    .eva_rd_addr(eva_rd_addr),
    .index_out(index_out),
    .inv_out(inv_out),
	  .block(block),
	  .done(done)
  );
 
  wire [`m-1:0] dout;	
	
  mem_dual #(.WIDTH(`m), .DEPTH(1 << `m), .FILE("P.in")) mem_dual_P (
	.clock(clk),
	.data_0(0),
	.data_1(0),
	.address_0(perm_rd_addr),
	.address_1(0),
	.wren_0(1'b0),
	.wren_1(1'b0),
	.q_0(perm_dout),
	.q_1(dout)
);
 
  wire [eva_mem_WIDTH-1:0] dout_1;

  mem_dual #(.WIDTH(eva_mem_WIDTH), .DEPTH(eva_mem_DEPTH), .FILE("eva.in")) mem_dual_eva (
	.clock(clk),
	.data_0(0),
	.data_1(0),
	.address_0(eva_rd_addr),
	.address_1(0),
	.wren_0(1'b0),
	.wren_1(1'b0),
	.q_0(eva_dout),
	.q_1(dout_1)
);
  
 initial
    begin
       $dumpfile("inv_g_alpha_tb.vcd");
       $dumpvars(0, inv_g_alpha_tb);
    end
  
  integer STDIN = 32'h8000_0000;
  integer STDERR = 32'h8000_0002;

  integer start_time;
  integer scan_file;
  
  initial
    begin
       scan_file = $fscanf(STDIN, "%b\n", block);
    end
  
  initial
    begin
		# 25;
		start_time = $time;
		start <= 1;
		# 10;
		start <= 0;
		@(posedge done);
		 $fdisplay(STDERR, "\nruntime: %0d cycles\n", ($time-start_time)/10);
		 $write("%b\n%b\n", index_out, inv_out);
		 # 50;
		 $finish;		
	  end

  always 
    #5 clk = ! clk; 
   
endmodule

