# This (sub-)makefile is part of the Classic McEliece hardware source code.
# 
# Copyright (C) 2022
#
# Authors: Norman Lahr <norman@lahr.email>
# 
# This source code is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# This source code is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this source code. If not, see <https://www.gnu.org/licenses/>. 

GEN_H_SRC_PATH := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

GEN_H_SUBMODULES := GEN_H_ENTRIES INV_G_ALPHA ADDITIVE_FFT

include $(GEN_H_SRC_PATH)/gen_H_entries/module.mk
include $(GEN_H_SRC_PATH)/inv_g_alpha/module.mk
include $(COMMON_SRC_PATH)/additive_fft/module.mk

GEN_H_SRC = gen_H.v

GEN_H_SRC += $(foreach module,$(GEN_H_SUBMODULES),$($(module)_SRC))

GEN_H =

gen_H: $(GEN_H)

# Specify targets for each parameter set
define GEN_H_SRC_TEMPLATE =

# Define all required targets to be included in the modules above
GEN_H_$(1) = $(addprefix $$(BUILD_DIR_SRC_$(1))/,$$(GEN_H_SRC))

# Collect all targets
GEN_H += GEN_H_$(1)

$$(BUILD_DIR_SRC_$(1))/gen_H.v: $$(GEN_H_SRC_PATH)/gen_H.v
	cp $$< $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call GEN_H_SRC_TEMPLATE,$(par))))
