# 
# Copyright (C) 2018
# function: generate theoretical result.
# Authors: Wen Wang <wen.wang.ww349@yale.edu>
#          Ruben Niederhagen <ruben@polycephaly.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import sys

import argparse

parser = argparse.ArgumentParser(description='Generate memory contents for H.',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
 
parser.add_argument('-t', dest='t', type=int, default=119,  
          help='number of errors; default = 119')
parser.add_argument('-N', dest='N', type=int, default=20,  
          help='column block size') 
parser.add_argument('-n', dest='n', type=int, default=8192,  
          help='code length') 
parser.add_argument('-m', '--gf', dest='m', type=int, default=13,
          help='field GF(2^m); default = GF(2^13)')
parser.add_argument('in_coeff', metavar='in_coeff', type=str,
                    help='input file containing coefficients')
parser.add_argument('in_P', metavar='in_P', type=str,
                    help='input file containing permuted indices')
 
args = parser.parse_args()

t = args.t
m = args.m
N = args.N
n = args.n
K = 2^m
L = m*t
K_ceil = N*int((n-1+N)/N)

F.<x> = GF(2)[]
KK.<a> = GF(2^m, modulus="first_lexicographic")
Kx.<x> = KK[]
 
fmt = "{0:0" + str(m) + "b}"

with open(args.in_coeff, "r") as f:
  coeff_in = f.readline().strip()

coeff = []

while coeff_in != "":
  coeff.append(Kx(KK.fetch_int(int(coeff_in[:m], 2))))
  coeff_in = coeff_in[m:]

# reconstruct the polynomial g(x)
g = Kx.zero()

for i, v in enumerate(reversed(coeff)):
  g += x^i*v

#print(g)

# reconstruct permutation indices
P_list = []

with open(args.in_P, "r") as f:
  for line in f:
    if not line.startswith("//"):
      if not (line == "\n"):
        P_list.append(int(line[:m], 2))

#print(len(P_list))
#print(P_list)

# polynomial evaluation
basis = [a^i for i in reversed(range(m))]

def convert(beta, num):
  ret = 0
  for i in range(len(beta)):
    if (num >> i) % 2:
      ret += beta[i]
  return ret

alpha = []

for j in range(2^m):
  alpha.append(KK(convert(basis, j)))

eva = []

for i in range(2^m):
  x = alpha[i]
  eva.append(g(x)) 
  if (g(x) == KK.zero()):
    print('ERROR!', i, x)
 
Y = matrix(KK, t, K)
for i in range(K):
  Y[0, i] = KK.one()
  for j in range(1, t):
    Y[j, i] = Y[j-1, i]*alpha[i]

#print(Y[1, 1])

Z = matrix(KK, K, K)
for i in range(K):
  Z[i, i] = (eva[i])^-1
     
P_trans = matrix(KK, K, K)
 
for i in range(K):
  row = P_list[i]
  P_trans[row, i] = 1

H = Y*Z*P_trans

#print(K_ceil)

H_expand = matrix(KK, t, K_ceil)
for i in range(K_ceil):
  for j in range(t):
    if (i < n):
      H_expand[j, i] = H[j, i]
    else:
      H_expand[j, i] = KK.zero()
 
for i in range(int(K_ceil/N)):
  for j in range(t):
    for k in reversed(range(m)):
      for q in reversed(range(N)):
        A = fmt.format(H_expand[j, N*i+q].integer_representation())
        sys.stdout.write(A[k])
      sys.stdout.write("\n")


