/*
 * This file is the testbench file for elim.v.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * M_elimERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, M_elimA 02110-1301  USA
 *
*/

`timescale 1ns / 1ps
module gen_H_tb;

  //inputs
  reg clk = 1'b0;
  reg rst = 1'b0;

  reg [1:0] op = 2'b11;

  reg pk_gen_start = 1'b0;
  reg eva_start = 1'b0;
  reg [`m*(`t+1)-1:0] eva_poly_in = 0;
  wire [`m-1:0] perm_dout;
  reg key_gen_done = 0;
   
  //outputs
  wire mem_wr_en;
  wire [`CLOG2(`L*K_ceil/`N) - 1 : 0] mem_wr_addr;
  wire [(`N*`M_elim)-1 : 0] mem_data_in;
  wire perm_rd_en;
  wire [`m-1:0] perm_rd_addr;
  wire init_done;
  wire eva_done;
  wire [2*`m*mem_width-1:0] eva_dout;
   
  localparam K_ceil = `N*((`N+`K-1)/`N);
  localparam mem_width = 1 << (`m-`CLOG2(`t)-1);
  
  gen_H #(.m(`m), .N(`N), .L(`L), .K(`K), .M_elim(`M_elim), .t(`t)) DUT (
   .clk(clk),
   .rst(1'b0),
   .op(op),
   .pk_gen_start(pk_gen_start),
   .eva_start(eva_start),
   .eva_poly_in(eva_poly_in),
   .mem_wr_en(mem_wr_en),
   .mem_wr_addr(mem_wr_addr),
   .mem_data_in(mem_data_in),
   .perm_rd_en(perm_rd_en),
   .perm_rd_addr(perm_rd_addr),
   .perm_dout(perm_dout),
   .init_done(init_done),
   .eva_done(eva_done),
   .eva_dout(eva_dout)
  );
  
 
// M_20k for storing dout
  mem #(.WIDTH(`N*`M_elim), .DEPTH(`L*K_ceil/`N), .FILE("")) mem_data (
  .clock (clk),
  .data (mem_data_in),
  .rdaddress (0),
  .rden (1'b0),
  .wraddress (mem_wr_addr),
  .wren (mem_wr_en),
  .q ()
);
  
// memory for storing permutation
  
  mem #(.WIDTH(`m), .DEPTH(1<<`m), .FILE("perm.in")) perm_mem_data (
    .clock (clk),
    .data (0),
    .rdaddress (perm_rd_addr),
    .rden (perm_rd_en),
    .wraddress (0),
    .wren (0),
    .q (perm_dout)
);
  
  initial
    begin
      $dumpfile("gen_H_tb.vcd");
      $dumpvars(0, gen_H_tb);
    end
 
  
  integer scan_file;
  
  integer STDIN  = 32'h8000_0000;

  initial
    begin
     scan_file = $fscanf(STDIN, "%b\n", eva_poly_in);
  end
  
  integer eva_start_time;
  integer eva_done_time;

  integer init_start_time;
  
  initial
    begin
      op <= 2'b01;
      # 25;
      eva_start <= 1;
      eva_start_time = $time;
      # 10;
      eva_start <= 0;
   end

  initial
    begin
      @(posedge eva_done);
      eva_done_time = $time;
      $display("\nruntime for polynomial evaluation: %0d cycles\n", ($time-eva_start_time)/10);
      $writememb("eva_A.out", DUT.add_FFT_inst.reduction_inst.mem_coeff_A.mem);
      $writememb("eva_B.out", DUT.add_FFT_inst.reduction_inst.mem_coeff_B.mem);
      # 10;
      op <= 2'b01;
      # 10;
      pk_gen_start <= 1'b1;
      init_start_time = $time;
      # 10;
      pk_gen_start <= 1'b0;
      @(posedge init_done);

      $display("first part done; gen second part - %0d cycles \n", ($time-init_start_time)/10);

      # 10;
      op <= 2'b10;
      # 10;
      pk_gen_start <= 1'b1;
      init_start_time = $time;
      # 10;
      pk_gen_start <= 1'b0;
      @(posedge init_done);

      $display("\ntotal runtime for init: %0d cycles  (%0d)\n", ($time-init_start_time)/10, ($time-eva_done_time)/10);
      $writememb("data.out", mem_data.mem);
      # 1000;
      $finish;
     end

  always 
    #5 clk = ! clk; 
   
endmodule
