/*
 * This file is the testbench file for elim.v.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * M_elimERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, M_elimA 02110-1301  USA
 *
*/

`timescale 1ns / 1ps
module gen_pk_tb;

  //inputs
  reg                                clk = 1'b0;
  reg                                rst = 1'b0;
  reg                                pk_gen_start = 1'b0;
  reg                                eva_start = 0;
  reg [`m*(`t+1)-1:0]                eva_poly_in = 0;
   
  reg                                rd_en = 0;
  reg [`CLOG2(`L*K_elim/`N) - 1 : 0] rd_addr = 0;
   
  wire [`m-1:0]                       perm_dout;
 
  //outputs
   
  wire init_done;
  wire eva_done;
  wire elim_done;
  
  wire perm_rd_en;
  wire [`m-1:0] perm_rd_addr;
  
  wire [`N*`M_elim-1:0] elim_data_out;
  wire [2*`m*mem_width-1:0] eva_dout;
  wire fail;
   
  localparam K_elim = `N*((`N+`K-1)/`N);
  localparam mem_width = 1 << (`m-`CLOG2(`t+1)-1); 
  
  gen_pk #(.m(`m), .N(`N), .L(`L), .K(`K), .M_elim(`M_elim), .t(`t), .BLOCK(`BLOCK)) DUT (
    .clk(clk),
    .rst(rst),
    .pk_gen_start(pk_gen_start),
    .eva_start(eva_start),
    .eva_poly_in(eva_poly_in),
    .rd_en(rd_en),
    .rd_addr(rd_addr),
    .init_done(init_done),
    .eva_done(eva_done),
    .elim_done(elim_done),
    .elim_data_out(elim_data_out),
    .perm_dout(perm_dout),
    .perm_rd_en(perm_rd_en),
    .perm_rd_addr(perm_rd_addr),
    .eva_dout(eva_dout),
    .fail(fail)
    );
  
  // memory for storing permutation
  
  mem #(.WIDTH(`m), .DEPTH(1<<`m), .FILE("perm.in")) perm_mem_data (
    .clock (clk),
    .data (0),
    .rdaddress (perm_rd_addr),
    .rden (perm_rd_en),
    .wraddress (0),
    .wren (0),
    .q (perm_dout)
);
 
  // initial
  //   begin
  //     $dumpfile("H_elim_tb.vcd");
	 //    $dumpvars(0, H_elim_tb);
  //   end
 
  
  integer scan_file, file;
	
  integer STDERR = 32'h8000_0002;
  integer STDIN  = 32'h8000_0000;

  initial
    begin
          file = $fopenr("data.in");
 	  scan_file = $fscanf(file, "%b\n", eva_poly_in);
          $fclose(file);
	end
	
  integer eva_start_time;
  integer eva_done_time;
  integer elim_start_time;
  
  initial
    begin
      rst <= 0;
      # 25;
      eva_start <= 1;
      eva_start_time = $time;
      # 10;
      eva_start <= 0;
      # 20;
       
	 end
  
  integer i;
  
  initial
	 begin
		@(posedge eva_done);
     eva_done_time = $time;
		  $display("\nruntime for polynomial evaluation: %0d cycles\n", ($time-eva_start_time)/10);
      $writememb("eva_A.out", DUT.gen_H_inst.add_FFT_inst.reduction_inst.mem_coeff_A.mem);
      $writememb("eva_B.out", DUT.gen_H_inst.add_FFT_inst.reduction_inst.mem_coeff_B.mem);
     # 20;
     pk_gen_start <= 1'b1;
     # 10;
     pk_gen_start <= 1'b0;
		@(posedge init_done);
      elim_start_time = $time;
		  $display("\ntotal runtime for init: %0d cycles\n", ($time-eva_done_time)/10);
      $writememb("init.out", DUT.systemizer_inst.phase_inst.step_inst.mem_data.mem);
     @(posedge elim_done, fail);
      $display("\ntotal runtime for elim part: %0d cycles\n", ($time-elim_start_time)/10);
      $writememb("mem_data.out", DUT.systemizer_inst.phase_inst.step_inst.mem_data.mem);
      $display("\ntotal runtime(add_FFT+init+elim): %0d cycles\n", ($time-eva_start_time)/10);
      # 20;
     // write output data
     @(negedge clk);
     rd_en <= 1'b1;
     file = $fopen("data.out");

     if (fail) begin
         $fwrite(file, "fail\n");
     end else begin
       for (i = 0; i < `L; i = i + 1)
         begin
           rd_addr <= (K_elim/`N-1)*`L+i;
           # 10;
           $fwrite(file, "%b\n", elim_data_out);
         end
     end
     $fclose(file);
		# 100000;
		$finish;
	 end

  always 
    #5 clk = ! clk; 
   
endmodule
