/*
 * This file is the testbench file for mem_init_R.v.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

`timescale 1ns / 1ps
module mem_init_R_tb;

  //inputs
  reg clk = 1'b0;
  reg start = 1'b0;
  reg [(`L*`M-1):0] coeff_in = 0;
 
  //outputs
  wire wr_en;
  wire [`CLOG2(`L*K_elim/`N) - 1 : 0] wr_addr;
  wire [(`N*`M)-1 : 0] data_in;
  wire init_done;
  
  localparam K_elim = `N*((`N+`K-1)/`N);
  
  mem_init_R #(.N(`N), .L(`L), .K(`K), .M(`M)) DUT (
    .clk(clk),
    .start(start),
    .coeff_in(coeff_in),
    .wr_en(wr_en),
    .wr_addr(wr_addr),
    .data_in(data_in),
    .init_done(init_done) 
  );
  
  
  // M20k for storing dout
    mem #(.WIDTH(`N*`M), .DEPTH(`L*K_elim/`N), .FILE("")) mem_data (
    .clock (clk),
    .data (data_in),
    .rdaddress ({`CLOG2(`L*K_elim/`N){1'b0}}),
    .rden (1'b0),
    .wraddress (wr_addr),
    .wren (wr_en),
    .q ()
  );
 
  //initial
  //  begin
  //   $dumpfile("mem_init_R_tb.vcd");
  //         $dumpvars(0, mem_init_R_tb);
  //  end
 
  
  integer scan_file;
  
  integer STDERR = 32'h8000_0002;
  integer STDIN  = 32'h8000_0000;

  initial
    begin
     scan_file = $fscanf(STDIN, "%b", coeff_in);
  end
  
  integer start_time;
  
  initial
    begin
    # 25;
    start_time = $time;
    start <= 1;
    # 10;
    start <= 0;
    @(posedge init_done);
      $writememb("data.out", mem_data.mem); 
      $fdisplay(STDERR, "\nruntime for initializing mem: %0d cycles\n", ($time-start_time)/10);
      
    # 50;
    $finish;
        
   end

  always 
    #5 clk = ! clk; 
   
endmodule

