# This (sub-)makefile is part of the Classic McEliece hardware source code.
# 
# Copyright (C) 2022
#
# Authors: Norman Lahr <norman@lahr.email>
# 
# This source code is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# This source code is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this source code. If not, see <https://www.gnu.org/licenses/>. 

GEN_POLY_SRC_PATH := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

GEN_POLY_SUBMODULES := MEM_INIT_R SYSTEMIZER_SPGFX_R_ELIM

include $(GEN_POLY_SRC_PATH)/mem_init_R/module.mk

### GFX R ELIM
#include $(COMMON_SRC_PATH)/systemizer/single_pass_GFx/module.mk

SYSTEMIZER_SPGFX_SRC_PATH = $(COMMON_SRC_PATH)/systemizer/single_pass_GFx/

SYSTEMIZER_SPGFX_R_ELIM_SUBMODULES := GF_MAD GF_INV LOG2 MEMORY_MEM DELAY

# Specific version (R_elim) for the generation of the irreducible polynomial
SYSTEMIZER_SPGFX_R_ELIM_SRC = elim_R_elim.v \
	phase_R_elim.v \
	step_R_elim.v \
	processor_B_R_elim.v \
	processor_AB_R_elim.v \
	comb_SA_R_elim.v

SYSTEMIZER_SPGFX_R_ELIM_SRC += $(foreach module,$(SYSTEMIZER_SPGFX_R_ELIM_SUBMODULES),$($(module)_SRC))

SYSTEMIZER_SPGFX_R_ELIM =

systemizer_spgfx_R_elim: $(SYSTEMIZER_SPGFX_R_ELIM)

# Specify targets for each parameter set
define SYSTEMIZER_SPGFX_R_ELIM_SRC_TEMPLATE =

SYSTEMIZER_SPGFX_R_ELIM_$(1) = $(addprefix $$(BUILD_DIR_SRC_$(1))/,$(SYSTEMIZER_SPGFX_R_ELIM_SRC))

SYSTEMIZER_SPGFX_R_ELIM += SYSTEMIZER_SPGFX_R_ELIM_$(1)

# Targets for the R_elim version

# $$(BUILD_DIR_SRC_$(1))/comb_SA_R_elim_pre.v: $$(SYSTEMIZER_SPGFX_SRC_PATH)/gen_comb_SA.py
# 	$$(PYTHON) $$< -n $(N_R_elim) --block $(BLOCK_R_elim) --gf $(m) --name comb_SA_R_elim > $$@

$$(BUILD_DIR_SRC_$(1))/comb_SA_R_elim.v: $$(SYSTEMIZER_SPGFX_SRC_PATH)/gen_comb_SA.py
	$$(PYTHON) $$< -n $(N_R_elim) --block $(BLOCK_R_elim) --gf $$(m_$(1)) --name comb_SA_R_elim > $$@
	sed -i 's/processor_AB/processor_AB_R_elim/g; s/processor_B/processor_B_R_elim/g' $$@

$$(BUILD_DIR_SRC_$(1))/processor_AB_R_elim.v: $$(SYSTEMIZER_SPGFX_SRC_PATH)/processor_AB.v
	sed 's/module processor_AB/module processor_AB_R_elim/g' $$< > $$@

$$(BUILD_DIR_SRC_$(1))/processor_B_R_elim.v: $$(SYSTEMIZER_SPGFX_SRC_PATH)/processor_B.v
	sed 's/module processor_B/module processor_B_R_elim/g' $$< > $$@

$$(BUILD_DIR_SRC_$(1))/step_R_elim.v: $$(SYSTEMIZER_SPGFX_SRC_PATH)/step.v
	sed 's/comb_SA #(.BLOCK(BLOCK)) comb_SA_inst/comb_SA_R_elim #(.BLOCK(BLOCK)) comb_SA_inst/g; s/module step/module step_R_elim/g' $$< > $$@

$$(BUILD_DIR_SRC_$(1))/phase_R_elim.v: $$(SYSTEMIZER_SPGFX_SRC_PATH)/phase.v
	sed 's/step #(.N(N), .M(M), .L(L), .K(K), .BLOCK(BLOCK), .DATA(DATA)) step_inst/step_R_elim #(.N(N), .M(M), .L(L), .K(K), .BLOCK(BLOCK), .DATA(DATA)) step_inst/g; s/module phase/module phase_R_elim/g' $$< > $$@

$$(BUILD_DIR_SRC_$(1))/elim_R_elim.v: $$(SYSTEMIZER_SPGFX_SRC_PATH)/systemizer.v
	sed 's/phase #(.N(N), .M(M), .L(L), .K(K), .BLOCK(BLOCK), .DATA(DATA)) phase_inst/phase_R_elim #(.N(N), .M(M), .L(L), .K(K), .BLOCK(BLOCK), .DATA(DATA)) phase_inst/g; s/module systemizer/module elim_R_elim/g' $$< > $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call SYSTEMIZER_SPGFX_R_ELIM_SRC_TEMPLATE,$(par))))

### end GFX R ELIM


GEN_POLY_SRC = gen_poly.v

GEN_POLY_SRC += $(foreach module,$(GEN_POLY_SUBMODULES),$($(module)_SRC))

GEN_POLY =

gen_poly: $(GEN_POLY)

# Specify targets for each parameter set
define GEN_POLY_SRC_TEMPLATE =

# Define all required targets to be included in the modules above
GEN_POLY_$(1) = $(addprefix $$(BUILD_DIR_SRC_$(1))/,$$(GEN_POLY_SRC))

# Collect all targets
GEN_POLY += GEN_POLY_$(1)

$$(BUILD_DIR_SRC_$(1))/gen_poly.v: $$(GEN_POLY_SRC_PATH)/gen_poly.v
	cp $$< $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call GEN_POLY_SRC_TEMPLATE,$(par))))
