# This file is the testbench for the gen_support module.
# 
# Copyright (C) 2018
# Authors: Wen Wang <wen.wang.ww349@yale.edu>
#          Ruben Niederhagen <ruben@polycephaly.org>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import argparse


parser = argparse.ArgumentParser(description='Generate random permutation P.',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('-m', '--gf', dest='m', type=int, default=10,
          help='field GF(2^m); default = GF(2^13)')

parser.add_argument('-t', dest='t', type=int, default=50,
          help='number of correctable errors')

parser.add_argument('-s1', '--sigma1', dest='sigma1', type=int, default=16,
          help='number of bits for poly')

args = parser.parse_args()

m = args.m
t = args.t

F2 = GF(2)
F2z.<z> = F2[]
F2zy.<y> = F2z[]

f = {
   9 : z^9+z^1+1, # for testing
  10 : z^10+z^3+1, # for testing
  12 : z^12+z^3+1,
  13 : z^13+z^4+z^3+z+1,
}

F = {
   20 : z^20+z^3+1, # for testing
   50 : y^50+y^5+y^2+z, # for testing
   64 : y^64+y^3+y+z,
   96 : y^96+y^10+y^9+y^6+1,
  119 : y^119+y^8+1,
  128 : y^128+y^7+y^2+y+1,
}

q = 2^m
sigma1 = args.sigma1


Fq.<zinFq> = GF(q,name='zinFq',modulus=f[m])
Fqt.<yinFqt> = Fq.extension(F[t])

## Fq.<zinFq> = GF(q,name='zinFq',modulus="first_lexicographic")
## g = PolynomialRing(GF(2), 'y').irreducible_element(t, algorithm="first_lexicographic")
## #g = PolynomialRing(Fq, 'y').irreducible_element(t, algorithm="random")
## 
## g = PolynomialRing(Fq, 'y')(g)
## print(g)
## print(g.is_irreducible())
## 
## Fqt.<yinFqt> = Fq.extension(g)

## Fq.<zinFq> = GF(q,name='zinFq',modulus="first_lexicographic")
## Fqt.<yinFqt> = Fq.extension(y^50+y^5+y^2+z)

Fqx.<x> = Fq[]

def is_allowed_polynomial(g):
  if parent(g) != Fqx: return False
  if not g.is_monic(): return False
  if not g.is_irreducible(): return False
  return g.degree() == t

def is_bit_vector(e,n):
  if len(e) != n: return False
  return all(parent(ei) == F2 for ei in e)

def Irreducible(d):
  assert is_bit_vector(d,sigma1*t)
  beta = [sum(d[sigma1*j+i]*zinFq^i for i in range(m)) for j in range(t)]
  assert all(parent(betaj) == Fq for betaj in beta)
  beta = sum(beta[j]*yinFqt^j for j in range(t))

  #print(d,beta)

  g = beta.minpoly()
  if g.degree() != t: return False
  assert is_allowed_polynomial(g)

  #print(g)

  return g


g = False

# Hardware module does not detect failure!
while not g:
  d = [F2.random_element() for i in range(sigma1*t)]
  
  g = Irreducible(d)

#print(len(d), sigma1, t)

#d = [d[i:i + sigma1] for i in range(0, len(d), sigma1)]
d = [d[i:i + m] for i in range(0, len(d), sigma1)]

# Need to reverse bit order!
#d_out = "".join(["".join(str(i) for i in v)[::-1] for v in d])
d_out = ["".join(str(i) for i in v)[::-1] for v in d]

#print(d_out)

d_out = "".join(d_out[::-1])

f = open("data.in", "w")
f.write(d_out)
#f.write("\n")
f.close()

fmt_str = "{0:0" + str(m) + "b}"

g_out = "\n".join([fmt_str.format(v.integer_representation()) for v in g][:-1])

#print(g_out)
 
f = open("elim.out", "w")
f.write(g_out)
f.write("\n")
f.close()

