
/*
 * This file is the testbench file for elim.v.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

`timescale 1ns / 1ps
module gen_poly_tb;

  //inputs
  reg clk = 1'b0;
  reg start = 1'b0;
  reg [(`L*`M-1):0] coeff_in = 0;
  reg rd_en = 1'b0;
  reg [`CLOG2(`L*K_elim/`N) - 1 : 0] rd_addr = 0;
  
  //outputs
  wire [(`N*`M)-1 : 0] data_out;
  wire done;
  
  localparam K_elim = `N*((`K+`N-1)/`N);
  
  gen_poly #(.N(`N), .L(`L), .K(`K), .M(`M), .BLOCK(`BLOCK)) DUT(
    .clk(clk),
    .rst(1'b0),
    .start(start),
    .rd_en(rd_en),
    .rd_addr(rd_addr),
    .init_done(),
    .data_out(data_out),
    .coeff_in(coeff_in),
    .done(done)
  );
 
  //initial
  //  begin
  //     $dumpfile("gen_poly_tb.vcd");
  //     $dumpvars(0, gen_poly_tb);
  //  end
 
  integer i;
  
  integer STDIN  = 32'h8000_0000;
  integer STDERR = 32'h8000_0002; 
  
  integer scan_file;

  initial
    begin
		scan_file = $fscanf(STDIN, "%b", coeff_in);
	 end
  
  integer start_time;
  integer end_time;

  initial
    begin
      start <= 1'b0;
      # 45;
      start <= 1'b1;
      start_time = $time;
      # 10;
      start <= 1'b0;
      @(posedge done);
      end_time = $time;
      $fdisplay(STDERR,"\nruntime: %0d cycles\n", (end_time - start_time)/10);    
      // write output data
      @(negedge clk);
      rd_en = 1'b1;
      for (i = `L*(K_elim/`N-1); i < `L*(K_elim/`N); i = i+1)
       begin
        rd_addr = i;
        # 10;
        $write("%b\n", data_out);
        //$write("%b\n", data_out[(`N*`M)-1:((`N-1)*`M)]);
        end
      $finish;
    end

  always 
    #5 clk <= ! clk; 
   
endmodule

