# This (sub-)makefile is part of the Classic McEliece hardware source code.
# 
# Copyright (C) 2022
#
# Authors: Norman Lahr <norman@lahr.email>
# 
# This source code is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# This source code is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this source code. If not, see <https://www.gnu.org/licenses/>. 

KEYGEN_SRC_PATH := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

TOPMODULE ?= keygen
KEYGEN_SUBMODULES := GEN_POLY FIELD_ORDERING GEN_PK

BUILD_DIR ?= $(KEYGEN_SRC_PATH)/build
RESULTS_DIR ?= $(BUILD_DIR)/results
COMMON_SRC_PATH ?= $(KEYGEN_SRC_PATH)/../common

include $(KEYGEN_SRC_PATH)/../../../config/parameters.mk

#TODO clean
PYTHON ?= python3
SAGE ?= sage
IVERILOG ?= iverilog
VERILATOR ?= verilator

.SECONDEXPANSION:

ifndef BUILD_DIR_TEMPLATE
define BUILD_DIR_TEMPLATE =

# Parameter-specific build directory root
BUILD_DIR_$(1) = $$(BUILD_DIR)/$$(TOPMODULE)/$(1)
# Parameter-specific source and testbench directory
BUILD_DIR_SRC_$(1) = $$(BUILD_DIR_$(1))/src
BUILD_DIR_TB_$(1) = $$(BUILD_DIR_$(1))/testbench
#BUILD_DIR_TB_DATA_$(1) = $$(BUILD_DIR_$(1))/testbench/data

endef

$(foreach par, $(PAR_SETS), $(eval $(call BUILD_DIR_TEMPLATE,$(par))))

endif


# Include the submodules after the directory definitions
include $(COMMON_SRC_PATH)/field_ordering/module.mk
include $(KEYGEN_SRC_PATH)/gen_sk/gen_poly/module.mk
include $(KEYGEN_SRC_PATH)/gen_pk/module.mk

# Set the source files of this module
KEYGEN_SRC = key_gen.v
# Add the sources of the submodules
KEYGEN_SRC += $(foreach module,$(KEYGEN_SUBMODULES),$($(module)_SRC))

# Define comulated target for generated sources
#.PHONY: sources
#sources: $(addprefix sources-, $(PAR_SETS))

KEYGEN =

#
# Define the sources targets
#
define KEYGEN_SRC_TEMPLATE =

# Define all required targets to be included in the modules above
KEYGEN_$(1) = $(addprefix $$(BUILD_DIR_SRC_$(1))/,$(KEYGEN_SRC))

# Collect all targets
KEYGEN += $$(KEYGEN_$(1))

# Define parameter-specific generated source targets
#.PHONY: sources-$(1)
#sources-$(1): $$(KEYGEN_$(1))

$$(BUILD_DIR_SRC_$(1)):
	mkdir -p $$@

$$(BUILD_DIR_SRC_$(1))/key_gen.v: $$(KEYGEN_SRC_PATH)/key_gen.v
	cp $$< $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call KEYGEN_SRC_TEMPLATE,$(par))))

# TODO Add testbench support
# # Set the testbench source and data files of this module
# KEYGEN_TB_SRC = decap_tb.v

# # Define comulated target for generated sources
# .PHONY: testbench
# testbench: $(addprefix testbench-, $(PAR_SETS))

# # Define comulated target for generated sources
# .PHONY: sim
# sim: $(addprefix sim-, $(PAR_SETS))

# KEYGEN_TB =

# #
# # Define the testbench targets
# #
# define KEYGEN_TB_TEMPLATE =

# KEYGEN_TB_DATA_$(1) = \
# 	C0_$$(id_$(1)).in \
# 	C1_$$(id_$(1)).in \
# 	P_$$(id_$(1)).in \
# 	poly_g_$$(id_$(1)).in \
# 	s_$$(id_$(1)).in

# # Define all required targets to be included in the modules above
# KEYGEN_TB_$(1) = $(addprefix $$(BUILD_DIR_TB_$(1))/,$(KEYGEN_TB_SRC) $$(KEYGEN_TB_DATA_$(1)))
# KEYGEN_TB_$(1) += $$(foreach data, $$(KEYGEN_TB_DATA_$(1)),$$(BUILD_DIR_TB_$(1))/$$(data))

# # Collect all targets
# KEYGEN_TB += $$(KEYGEN_TB_$(1))

# # Define parameter-specific generated testbench building targets
# .PHONY: testbench-$(1)
# testbench-$(1): sources-$(1) $$(KEYGEN_TB_$(1))

# $$(BUILD_DIR_TB_$(1)):
# 	mkdir -p $$@

# $$(BUILD_DIR_TB_$(1))/decap_tb.v: $$(KEYGEN_SRC_PATH)/testbench/decap_tb.v
# 	cp $$< $$@

# $$(BUILD_DIR_TB_$(1))/%.in: $$(KEYGEN_SRC_PATH)/testbench/data/%.in
# 	cp $$< $$@


# # Define parameter-specific generated testbench run targets
# .PHONY: sim-$(1)
# sim-$(1): run_decapsualtion_$(1)_sim_iverilog

# # Build iverilog binary
# # Note: The macro definition file must be at the very beginning.
# $$(BUILD_DIR_TB_$(1))/decapsualtion_sim_iverilog: testbench-$(1)\
# 	| $$(RESULTS_DIR)
# 	$(IVERILOG) -Wall -Wno-timescale \
# 	-Dparameter_set=$$(id_$(1)) \
# 	-Dmem_width=$$(mem_width_$(1)) -DBLOCK=$$(BLOCK) \
# 	-Dmul_sec_BM=$$(mul_sec_BM) -Dmul_sec_BM_step=$$(mul_sec_BM_step) \
# 	-DFILE_MEM_CONSTS=\"$$(BUILD_DIR_SRC_$(1))/mem_consts\" \
# 	-DFILE_P=\"$$(BUILD_DIR_TB_$(1))/P.in\" \
# 	-DFILE_POLYG=\"$$(BUILD_DIR_TB_$(1))/poly_g.in\" \
# 	-DFILE_CIPHER=\"$$(BUILD_DIR_TB_$(1))/cipher.in\" \
# 	-DFILE_DATA_OUT=\"$$(BUILD_DIR_TB_$(1))/data.out\" \
# 	-DFILE_RE_ENC_OUT=\"$$(BUILD_DIR_TB_$(1))/re_encryption_res.out\" \
# 	-DFILE_DOUBLED_SYNDROME_OUT=\"$$(BUILD_DIR_TB_$(1))/doubled_syndrome.out\" \
# 	-DFILE_VCD=\"$$(BUILD_DIR_TB_$(1))/wave.vcd\" \
# 	-DFILE_CYCLES_PROFILE=\"$$(RESULTS_DIR)/cycles_profile_$(1).data\" \
# 	-s decap_tb \
# 	-I $$(BUILD_DIR_SRC_$(1)) \
# 	$$(BUILD_DIR_SRC_$(1))/clog2.v \
# 	$$(BUILD_DIR_TB_$(1))/decap_tb.v \
# 	$$(filter-out $$(BUILD_DIR_SRC_$(1))/mem_consts $$(BUILD_DIR_SRC_$(1))/clog2.v,$$(sort $$(KEYGEN_$(1)))) \
# 	-o $$@

# # Run the iverilog binary
# run_decapsualtion_$(1)_sim_iverilog: $$(BUILD_DIR_TB_$(1))/decapsualtion_sim_iverilog
# 	$$<
# 	# @diff $$(BUILD_DIR_TB_$(1))/data.out $$(BUILD_DIR_TB_$(1))/error.in \
# 	# && echo "Test Passed!"

# $$(BUILD_DIR_TB_$(1))/Vkeygen_tb: $$(KEYGEN_SRC_PATH)/testbench/tb.cpp testbench-$(1) | $$(RESULTS_DIR)
# 	verilator \
# 	--unroll-count 1024 \
# 	-Wall -Wno-fatal \
# 	--trace \
# 	--Mdir $$(BUILD_DIR_TB_$(1))/obj_dir \
# 	--top-module decap_tb \
# 	-Gparameter_set=$$(id_$(1)) \
# 	-DFILE_MEM_CONSTS=\"$$(BUILD_DIR_SRC_$(1))/mem_consts\" \
# 	-DFILE_C0=\"$$(BUILD_DIR_TB_$(1))/C0_$$(id_$(1)).in\" \
# 	-DFILE_C1=\"$$(BUILD_DIR_TB_$(1))/C1_$$(id_$(1)).in\" \
# 	-DFILE_P=\"$$(BUILD_DIR_TB_$(1))/P_$$(id_$(1)).in\" \
# 	-DFILE_S=\"$$(BUILD_DIR_TB_$(1))/s_$$(id_$(1)).in\" \
# 	-DFILE_POLYG=\"$$(BUILD_DIR_TB_$(1))/poly_g_$$(id_$(1)).in\" \
# 	-DFILE_ERROR_OUT=\"$$(BUILD_DIR_TB_$(1))/error.out\" \
# 	-DFILE_RE_ENC_OUT=\"$$(BUILD_DIR_TB_$(1))/re_encryption_res.out\" \
# 	-DFILE_DOUBLED_SYNDROME_OUT=\"$$(BUILD_DIR_TB_$(1))/doubled_syndrome.out\" \
# 	-DFILE_K_OUT=\"$$(BUILD_DIR_TB_$(1))/K_$$(id_$(1)).out\" \
# 	-DFILE_VCD=\"$$(BUILD_DIR_TB_$(1))/wave.vcd\" \
# 	-DFILE_CYCLES_PROFILE=\"$$(RESULTS_DIR)/cycles_profile_$(1).data\" \
# 	-I$$(BUILD_DIR_SRC_$(1)) \
# 	-I$$(BUILD_DIR_TB_$(1)) \
# 	--cc \
# 	$$(BUILD_DIR_SRC_$(1))/clog2.v \
# 	$$(BUILD_DIR_TB_$(1))/decap_tb.v \
# 	$$(filter-out $$(BUILD_DIR_SRC_$(1))/mem_consts $$(BUILD_DIR_SRC_$(1))/clog2.v,$$(sort $$(KEYGEN_$(1)))) \
# 	--exe $$< \
# 	--build \
# 	-j 8 \
# 	--threads 8 \
# 	-o $$@

# # Run the iverilog binary
# run_decapsualtion_$(1)_sim_verilator: $$(BUILD_DIR_TB_$(1))/Vdecap_tb
# 	$$<
# 	# @diff $$(BUILD_DIR_TB_$(1))/data.out $$(BUILD_DIR_TB_$(1))/error.in \
# 	# && echo "Test Passed!"

# endef

# $(foreach par, $(PAR_SETS), $(eval $(call KEYGEN_TB_TEMPLATE,$(par))))

# Add the build directory as order-only-prerequisites to every source file target
$(KEYGEN) $(KEYGEN_TB): | $$(@D)

$(RESULTS_DIR):
	mkdir -p $@

# clean_decryption_src_$(1):
# 	rm -rf $$(DEC_SRC_$(1))

# clean_decryption_src: $(addprefix clean_, $(BUILD_KEYGEN_SRC))
