/*
 * This file is the testbench file for key_gen.v.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

//`timescale 1ns/1ps

module key_gen_tb (
    input wire clk,
    output reg done
  );

//  initial
//    begin
//      $dumpfile("obj_dir/key_gen_tb.vcd");
//      $dumpvars(0, key_gen_tb);
//    end

  reg init_start = 1'b1;
  always @(posedge clk)
    init_start <= 1'b0;
 
  reg [1:0] request = 1'b0;
  
  reg host_sk_poly_rd_en = 1'b0;
  reg [`CLOG2(`T*K_R_elim/`N_R)-1:0]   host_sk_poly_rd_addr = 0;
  wire [`M*`N_R-1:0] sk_poly_dout;
  wire sk_poly_gen_done;
  
  reg                  host_beta_rand_wr_en;
  reg [`CLOG2(`T)-1:0] host_beta_rand_wr_addr;
  wire [`SIGMA1-1:0]    host_beta_rand_data_in;

  reg                  host_alpha_rand_wr_en;
  reg [`M-1:0]         host_alpha_rand_wr_addr;
  wire [`SIGMA2-1:0]    host_alpha_rand_data_in;

  //reg [63:0] seed_1 = 0;
  reg host_perm_rd_en = 1'b0;
  reg [`M-1:0] host_perm_rd_addr = 0;
  wire [`M-1:0] perm_dout;
  wire perm_done;
  
  //reg host_pk_rd_en = 0;
  //reg [`CLOG2(L*K_H_elim/`N_H)-1:0] host_pk_rd_addr = 0;
  //wire [`N_H-1:0] pk_dout;
  wire pk_gen_done;
  wire pk_gen_fail;
 
  parameter L = `M*`T;
  parameter K_H_elim = `N_H*((`K+`N_H-1)/`N_H);
  parameter K_R_elim = `N_R*((`T+1+`N_R-1)/`N_R); 
  
  key_gen #(.M(`M), .T(`T), .K(`K), .N_H(`N_H), .BLOCK_H(`BLOCK_H), .N_R(`N_R), .BLOCK_R(`BLOCK_R), .SIGMA1(`SIGMA1), .SIGMA2(`SIGMA2), .MEM_WIDTH(`MEM_WIDTH)) DUT (
    .clk(clk),
    .start(init_done || restart),
    .request(request),
    //
    .host_beta_rand_wr_en(host_beta_rand_wr_en),
    .host_beta_rand_wr_addr(host_beta_rand_wr_addr),
    .host_beta_rand_data_in(host_beta_rand_data_in),
    //
    .host_sk_poly_rd_en(host_sk_poly_rd_en),
    .host_sk_poly_rd_addr(host_sk_poly_rd_addr),
    .sk_poly_dout(sk_poly_dout),
    .sk_poly_gen_done(sk_poly_gen_done),
    //
    .host_alpha_rand_wr_en(host_alpha_rand_wr_en),
    .host_alpha_rand_wr_addr(host_alpha_rand_wr_addr),
    .host_alpha_rand_data_in(host_alpha_rand_data_in),
    //
    .host_perm_rd_en(host_perm_rd_en),
    .host_perm_rd_addr(host_perm_rd_addr),
    .perm_dout(perm_dout),
    .perm_done(perm_done),
    .host_pk_rd_en(host_pk_rd_en),
    .host_pk_rd_addr(host_pk_rd_addr),
    .pk_dout(pk_dout),
    .pk_gen_done(pk_gen_done),
    .pk_gen_fail(pk_gen_fail)
   
  );

  integer data_file, scan_file;

  integer data;
  integer i;

  reg [`SIGMA2-1:0] rand_alpha_data [0:(1<<`M)];
  reg [`SIGMA1-1:0] rand_beta_data [0:`T - 1];

  initial
    begin
      data_file = $fopen("sage_alpha_rand.out", "r");
      i = 0;
      while (!$feof(data_file)) begin
         scan_file = $fscanf(data_file, "%b\n", data);
         rand_alpha_data[i] = data;
         i = i + 1;
      end
      $fclose(data_file);

      data_file = $fopen("sage_beta_rand.out", "r");
      i = 0;
      while (!$feof(data_file)) begin
         scan_file = $fscanf(data_file, "%b\n", data);
         rand_beta_data[i] = data;
         i = i + 1;
      end
      $fclose(data_file);

      host_alpha_rand_wr_addr <= 0;
      host_beta_rand_wr_addr <= 0;
    end


  wire init_alpha_done;
  reg init_alpha_done_reg = 1'b0;

  assign init_alpha_done = host_alpha_rand_wr_addr == {`M{1'b1}};

  assign host_alpha_rand_data_in = rand_alpha_data[host_alpha_rand_wr_addr];

  reg do_restart = 1'b1;
  reg restart = 1'b0;


  always @(posedge clk)
    begin
      restart <= pk_gen_fail ? do_restart : 1'b0;
      do_restart <= pk_gen_fail ? 1'b0 : do_restart;

      init_alpha_done_reg <= init_alpha_done ? 1'b1 : init_done ? 1'b0 : init_alpha_done_reg;

      host_alpha_rand_wr_en <= init_start ? 1'b1 :
                               init_alpha_done ? 1'b0 : host_alpha_rand_wr_en;

      host_alpha_rand_wr_addr <= init_start ? 0 :
                                 host_alpha_rand_wr_en ? (host_alpha_rand_wr_addr + 1) : host_alpha_rand_wr_addr;
    end


  wire init_beta_done;
  reg init_beta_done_reg = 1'b0;

  assign init_beta_done = host_beta_rand_wr_addr == `T-1;

  assign host_beta_rand_data_in = rand_beta_data[host_beta_rand_wr_addr];

  always @(posedge clk)
    begin
      init_beta_done_reg <= init_beta_done ? 1'b1 : init_done ? 1'b0 : init_beta_done_reg;

      host_beta_rand_wr_en <= init_start ? 1'b1 :
                              init_beta_done ? 1'b0 : host_beta_rand_wr_en;

      host_beta_rand_wr_addr <= init_start ? 0 :
                                host_beta_rand_wr_en ? (host_beta_rand_wr_addr + 1) : host_beta_rand_wr_addr;
    end

  assign init_done = init_alpha_done_reg && init_beta_done_reg;
    
  integer sk_gen_start_time;

  integer sk_poly_gen_done_time;
  integer sk_perm_gen_done_time;

  integer pk_gen_start_time;
  
  always @(posedge DUT.sk_gen_start)
    begin
      sk_gen_start_time = $time;
    end

  always @(posedge DUT.sk_poly_gen_done)
    begin
      sk_poly_gen_done_time = $time;
      $display("\n running time for generating g(x): %0d cycles", ($time-sk_gen_start_time)/10);   
    end

  always @(posedge DUT.perm_done)
    begin
      sk_perm_gen_done_time = $time;
      $display("\n running time for permutation: %0d cycles", ($time-sk_gen_start_time)/10);   
    end
  
  
  always @(posedge DUT.pk_gen_start)
   begin
      pk_gen_start_time = $time;
      $display("\n running time for generating secret key: %0d cycles", ($time-sk_gen_start_time)/10);   
   end

  integer pk_gen_right_start_time = -1;

  always @(posedge DUT.pk_gen_inst.success)
   begin
      pk_gen_right_start_time = $time;
      $display("\n running time for checking left part: %0d cycles", ($time-pk_gen_start_time)/10);   
   end

  always @(posedge DUT.pk_gen_inst.elim_done)
   begin
     if (pk_gen_right_start_time > 1)
      $display("\n running time for computing right part: %0d cycles", ($time-pk_gen_right_start_time)/10);   
   end


  always  @(posedge clk)
    begin
      done <= pk_gen_done | (!do_restart && pk_gen_fail);
    end

  //// debug output
  //reg [`M-1:0] pk_perm_rd_addr_delay = 0;
  //reg pk_perm_rd_en_delay = 1'b0;

  //always 
  //  @(posedge clk)
  //  begin
  //    pk_perm_rd_addr_delay <= DUT.pk_perm_rd_addr;
  //    pk_perm_rd_en_delay <= DUT.pk_perm_rd_en;

  //    if (DUT.buf_sk_poly_done) begin
  //      $display("%b", {DUT.eva_coeff_in_one, DUT.buf_sk_poly});
  //    end

  //    if (pk_perm_rd_en_delay) begin
  //      $display("%d %b", pk_perm_rd_addr_delay, DUT.perm_dout);
  //    end
  //  end

  always 
    @(posedge pk_gen_fail)
    begin
      $display("\nkey gen failed!\n\n");
      //$finish();
    end

  integer f2, j, k;
  
  always 
    @(posedge pk_gen_done)
    begin
      $display("\n running time for generating public key: %0d cycles", ($time-pk_gen_start_time)/10);    
      $display("\n running time for generating a key pair: %0d cycles", ($time-sk_gen_start_time)/10);   

      f2 = $fopen("pk.out","w");
      
      for (i = 0; i < L; i = i+1)
        begin
          for (k = i + L*(L/`N_H); k < L*(K_H_elim/`N_H) - L; k = k+L)
            begin
              for (j = 0; j < `N_H; j = j+1)
                $fwrite(f2, "%b", DUT.pk_gen_inst.systemizer_inst.phase_inst.step_inst.mem_data.mem[k][j]);
            end
          for (j = 0; j < `N_H - (K_H_elim - `K); j = j+1)
            $fwrite(f2, "%b", DUT.pk_gen_inst.systemizer_inst.phase_inst.step_inst.mem_data.mem[k][j]);
          $fwrite(f2, "\n");
        end
      $fclose(f2);
      //$finish();
   end
  
endmodule

